# Lint as: python2, python3
# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
r"""Utilities for creating TFRecords of TF examples for the Open Images dataset.
"""

import tensorflow as tf


def open_sharded_output_tfrecords(exit_stack, base_path, num_shards):
    """Opens all TFRecord shards for writing and adds them to an exit stack.
    Args:
      exit_stack: A context2.ExitStack used to automatically closed the TFRecords
        opened in this function.
      base_path: The base path for all shards
      num_shards: The number of shards
    Returns:
      The list of opened TFRecords. Position k in the list corresponds to shard k.
    """

    if num_shards > 1:
        tf_record_output_filenames = [
            '{}-{:05d}-of-{:05d}'.format(base_path, idx, num_shards)
            for idx in range(num_shards)
        ]
    else:
        tf_record_output_filenames = [base_path]

    tfrecords = [
        exit_stack.enter_context(tf.io.TFRecordWriter(file_name))
        for file_name in tf_record_output_filenames
    ]

    return tfrecords



from typing import List, Dict, Tuple


def id_and_visit_from_filename(filename: str) -> Tuple[int, int]:
    """ Extract the SOF ID and the visit from a given DICOM filename
    :param filename: filenamne of a .dcm file
    :return: (id: int, visit: int) SOF id and visit as encoded in the filename
    """
    import re
    pattern = "SF([0-9]+)V([0-9]+)H.dcm"
    match = re.match(pattern, filename)
    return int(match.groups()[0]), int(match.groups()[1])


def read_label_file(filename: str) -> List[Dict]:
    """ Reads proximal femur detection labels from a label-studio json file and returns
        it's content in a cleaned up way.
    :param filename: Path to a label-studio json file
    :return: List of dictionary containing the information:
        (id: str, flipped: int, left_incomplete: int, left_incomplete_shaft:int, left_implant,
         right_incomplete: int, right_incomplete_shaft:int, right_implant,
         left_bbox_x: float, left_bbox_y: float, left_bbox_w: float, left_bbox_h: float,
         right_bbox_x: float, right_bbox_y: float, right_bbox_w: float, right_bbox_h: float) where the id
         is stored in the format <SOF_ID>V<Visit> and all int values are either 0 (false/absent) or 1 (true/present).
         The bbox coordinates are given in absolute pixel coordinates.
    """
    import json
    import re
    from pathlib import Path
    from zipfile import ZipFile

    labels = []

    label_path = Path(filename)

    open_file = lambda: open(filename, 'r')

    skipped = []

    if label_path.suffix == '.zip':
        zip_file = ZipFile(label_path, 'r')
        open_file = zip_file.open('result.json', 'r')

    with open_file() as fh:
        for item in json.load(fh):
            matches = re.match('^.*[/-]?([a-zA-B]*[0-9]+)V([0-9])+(L|R)-([0-9]+)x([0-9]+)\.png$', item['image']).groups()
            image_filename = re.match('^(.*/)?([^/]+\.png)$', item['image']).groups()[-1]
            if not matches or len(matches) != 5:
                raise ValueError(f"Annotated image '{item['image']}' has wrong format!")

            image_id = int(matches[0])
            visit = int(matches[1])
            lr = matches[2]
            width = int(matches[3])
            height = int(matches[4])

            if "labels" in item:
                image_labels = item['labels'] if isinstance(item['labels'], List) else [item['labels']]
            else:
                image_labels = []
            incomplete = 1 if "Incomplete" in image_labels else 0
            implant = 1 if "Implant" in image_labels else 0
            upside_down = 1 if "UpsideDown" in image_labels else 0

            invalid_keypoints = (incomplete + implant) > 0

            if not invalid_keypoints and ("keypoints" not in item or len(item['keypoints']) != 12):
                import sys
                print(f"Annotation for {image_filename} has invalid keypoints. Skipping.", file=sys.stderr)
                skipped.append(image_filename)
                continue

            entry = {
                'filename': image_filename,
                'id': image_id,
                'visit': visit,
                'left_right': lr,
                'upside_down': upside_down,
                'incomplete': incomplete,
                'implant': implant,
                'width': width,
                'height': height
            }

            keypoints = item['keypoints'] if not invalid_keypoints else [i for i in range(12)]

            bbox_min_x = min([float(kp['x']) for kp in keypoints]) if not invalid_keypoints else 0.0
            bbox_max_x = max([float(kp['x']) for kp in keypoints]) if not invalid_keypoints else 100.0
            bbox_min_y = min([float(kp['y']) for kp in keypoints]) if not invalid_keypoints else 0.0
            bbox_max_y = max([float(kp['y']) for kp in keypoints]) if not invalid_keypoints else 100.0

            entry['bbox_min_x'] = bbox_min_x
            entry['bbox_max_x'] = bbox_max_x
            entry['bbox_min_y'] = bbox_min_y
            entry['bbox_max_y'] = bbox_max_y

            for index, kp in enumerate(keypoints):
                entry[f"keypoint_x_{index}"] = float(kp['x']) if not invalid_keypoints else 0.0
                entry[f"keypoint_y_{index}"] = float(kp['y']) if not invalid_keypoints else 0.0

            labels.append(entry)

    return labels, skipped



""" Collection of dicom related functions.
"""

import pathlib
from typing import Tuple, Generator

import numpy as np


def read_meta(dcm_filename: str) -> Tuple[int, int, int, int]:
    """ Reads the meta data from the DICOM file and returns the width, height
    and the minimum and maximum pixel values.
    :param dcm_filename: path to DICOM file to read meta data from
    :return: (width, height, minPixelValue, maxPixelValue)
    """
    from pydicom import dcmread
    dcm = dcmread(dcm_filename, stop_before_pixels=True)
    return dcm.Columns, dcm.Rows, dcm.SmallestImagePixelValue, dcm.LargestImagePixelValue


def list_files(dir: str, recursive: bool = False) -> Generator[pathlib.Path, None, None]:
    """ List all DICOM files the the given directory, optionally recursively.
    :param dir: Path to a directory to look for DICOM files in
    :param recursive: If true: search in subdirectories
    :return: Generator over pathlib.Path objects
    """
    path = pathlib.Path(dir)
    pattern = '**/*.dcm' if recursive else '*.dcm'
    for p in path.glob(pattern):
        yield p


def list_meta(dir: str, recursive: bool = False) -> Generator[Tuple[str, int, int, int, int], None, None]:
    """
    List meta data (width, size, min and max pixel values) for all DICOM files
    inside the given directory (optionally including sub directories).
    :param dir: Path to a directory to look for DICOM files in
    :param recursive: If true: search in subdirectories
    :return: Generator of Tuples of (filename, width, height, minPixelValue, maxPixelValue)
    """
    for file in list_files(dir, recursive):
        meta = (file.name, *read_meta(str(file)))
        yield meta


def is_corrupted(dcm_filename: str) -> bool:
    """
    Checks if the given DICOM file is corrupted ot not, i.e. if the data array can be read or not/
    :param fcm_filename: path to the DICOM file to check for corruption
    :return: True iff the data array could NOT be read
    """
    from pydicom import dcmread
    import numpy as np
    try:
        dcm = dcmread(dcm_filename, stop_before_pixels=False)
        # access pixel array
        _ = dcm.pixel_array.astype(np.int16)
        return False
    except:
        return True


def find_corrupted(dir: str, recursive: bool = False) -> Generator[Tuple[bool, str], None, None]:
    """
    List all corrupted DICOM files inside the given directory (optionally including sub directories).
    :param dir: Path to a directory to look for corrupted DICOM files in
    :param recursive: If True: search in subdirectories
    :return: Generator of (is_corrupted, filename) pairs.
    """
    for file in list_files(dir, recursive):
        yield is_corrupted(str(file)), file.name


def read_image(dcm_filename: str) -> np.array:
    """ Read the pixel data from the given DICOM file
    :param dcm_filename: path to DICOM file to read the pixel data from
    :return: numpy array containing the pixel data
    """
    from pydicom import dcmread

    dcm = dcmread(dcm_filename)
    return dcm.pixel_array



from typing import Tuple, Union, List, Set, Dict

import tensorflow as tf


def export_images(dataset: tf.data.Dataset,
                  target_path: str,
                  format: str = 'png',
                  downsample_to: Tuple[Union[int, None], Union[int, None]] = (1, None),
                  split_lr: bool = False,
                  flip_lr: bool = False,
                  visits: List[int] = [],
                  included_ids: Set[int] = set(),
                  excluded_ids: Set[int] = set(),
                  num_groups: Union[None, int] = None,
                  max_group_size: Union[None, int] = None,
                  randomized_groups: bool = False,
                  zip: bool = False):
    """ Export the given dataset to png images
    :param dataset: Dataset to export
    :param target_path: path where the images will be exported to.
    :param format: image format to use: png (default) or jpeg
    :param downsample_to: target (width, height) of exported image (defaults to (None, None)).
        If one entry is None, it is inferred from the other one by keeping the aspect ratio.
        If booth entries are None, the original size is kept.
    :split_lr: if true (default is false), splits the image vertically into a left and a right part, i.e. two instead of
        one image files are written per example.
    :flip_lr: if true (default is false), flips the right part of the image vertically. Does nothing if
        'split_lr == False'.
    :visits: list of visits to include in the export. If visits is empty all visits are exported.
    :included_ids: set of SOF IDs to include. If empty, all IDs are exported.
    :excluded_ids: set of SOF IDs that should be excluded from the export.
    :num_groups: number of groups the exported files should be split into. If both, `max_group_size` and `num_groups` are `None` no grouping will be performed.
    :max_group_size: maximum number of files per group. If both, `max_group_size` and `num_groups` are `None` no grouping will be performed.
    :randomized_groups: If true, file to group assignments will be random.
    :zip: If true, write files into a zip file instead of a directory. If grouping is enabled, create one zip file per group.
    """
    from pathlib import Path
    from tqdm import tqdm
    from math import ceil
    from zipfile import ZipFile

    groups = {}
    group_sizes = []
    if num_groups or max_group_size:
        groups, group_sizes = _group_items(dataset,
                                           num_groups=num_groups,
                                           max_group_size=max_group_size,
                                           randomized=randomized_groups,
                                           split_lr=split_lr,
                                           visits=visits,
                                           included_ids=included_ids,
                                           excluded_ids=excluded_ids)

    target_path = Path(target_path)
    if zip and not groups:
        target_path.parent.mkdir(parents=True, exist_ok=True)
    else:
        target_path.mkdir(parents=True, exist_ok=True)

    if format.lower() == 'png':
        encoding_func = lambda x: tf.io.encode_png(x)
    elif format.lower() == 'jpeg':
        encoding_func = lambda x: tf.io.encode_jpeg(x)
    else:
        raise ValueError(f"Unsupported image format: {format}. supported formats are: png, jpeg")

    image_shape = next(iter(dataset))['image'].shape

    # Compute missing downsample_to entry, if any
    if not downsample_to[0] and not downsample_to[1]:
        downsample_to = (image_shape[1], image_shape[0])
    elif not downsample_to[0]:
        ratio = float(downsample_to[1]) / float(image_shape[0])
        downsample_to = (int(ceil(ratio * float(image_shape[0]))), downsample_to[1])
    elif not downsample_to[1]:
        ratio = float(downsample_to[0]) / float(image_shape[1])
        downsample_to = (downsample_to[0], int(ceil(ratio * float(image_shape[1]))))

    def group_prefix(sof_id, visit, lr):
        if not groups:
            return ''
        group = groups[(sof_id, visit, lr)]
        group_size = group_sizes[group]

        return f"G{group:04d}-N{group_size}-"

    if zip and not groups:
        zip_file = ZipFile(target_path, 'w')
    else:
        zip_files = [ZipFile(target_path.joinpath(f"{group:04d}-N{group_sizes[group]}.zip"), 'w') for group in
                     range(len(group_sizes))]

    def write_file(filename, bytes, group=None):
        if not zip:
            with open(str(filename), 'wb') as f:
                f.write(bytes)
        elif not groups:
            zip_file.writestr(filename.name, bytes)
        else:
            zf = zip_files[group]
            zf.writestr(filename.name, bytes)

    for example in tqdm(dataset):
        if "id" in example:
            sof_id = example['id'].numpy()
            visit = example['visit'].numpy()
        else:
            sof_id = example['image/id'].numpy()
            visit = example['image/visit'].numpy()
        if visits and visit not in visits:
            continue
        if included_ids and sof_id not in included_ids or sof_id in excluded_ids:
            continue

        image = tf.cast(tf.image.resize(example['image'], (downsample_to[1], downsample_to[0])), dtype=tf.uint8)

        if split_lr:
            left_img, right_img = split_image_lr(image, flip_lr)
            postfix = f"-{left_img.shape[1]}x{left_img.shape[0]}"
            # write left image, ie.e right hip
            key = (sof_id, visit, 'R')
            group = groups[key] if groups else None
            pref = group_prefix(*key)
            filename = target_path.joinpath(f'{pref}{sof_id}V{visit}R{postfix}.png')
            encoded_image = encoding_func(left_img)
            write_file(filename, encoded_image.numpy(), group)

            # write right image, i.e. left hip
            key = (sof_id, visit, 'L')
            group = groups[key] if groups else None
            pref = group_prefix(*key)
            filename = target_path.joinpath(f'{pref}{sof_id}V{visit}L{postfix}.png')
            encoded_image = encoding_func(right_img)
            write_file(filename, encoded_image.numpy(), group)
        else:
            postfix = f"-{image.shape[1]}x{image.shape[0]}"
            key = (sof_id, visit, '')
            group = groups[key] if groups else None
            pref = group_prefix(*key)
            filename = target_path.joinpath(f'{pref}{sof_id}V{visit}{postfix}.png')
            encoded_image = encoding_func(image)
            write_file(filename, encoded_image.numpy(), group)


def split_image_lr(image: tf.Tensor, flip_lr: bool = False) -> Tuple[tf.Tensor, tf.Tensor]:
    """ Split image vertically into a left and a right image.
    If the image does not have an even width, it is padded by one column at the right.
    :param image: source image to split
    :param flip_lr: If true, flips the right image vertically.
    :return: (left_image, right_image) tuple containing the left and the right half of the image.
    """
    padding = image.shape[1] % 2
    padded_image = tf.image.pad_to_bounding_box(image, 0, 0, image.shape[0], image.shape[1] + padding)

    image_half_width = tf.cast(tf.math.ceil(padded_image.shape[1] / 2), tf.int64)

    left_image = padded_image[..., 0:image_half_width, :]
    right_image = padded_image[..., image_half_width:, :]

    if flip_lr:
        right_image = tf.image.flip_left_right(right_image)

    return left_image, right_image


def _group_items(dataset: tf.data.Dataset,
                 max_group_size: Union[None, int],
                 num_groups: Union[None, int],
                 randomized: bool,
                 split_lr: bool,
                 visits: List[int],
                 included_ids: Set[int],
                 excluded_ids: Set[int]) -> Tuple[Dict[Tuple[int, int, str], int], List[int]]:
    """ Assigns each to be exported items to a group
    :param dataset: Dataset
    :param max_group_size: Maximum group size, only when num_groups is not given
    :param num_groups: Maximum number of groups, only when max_group_size is not given
    :param randomized: If true, assigment will be random
    :param split_lr: Tf example images should be split in left and right images
    :param visits: visits to include, if empty all visits are included
    :param included_ids: list of SOF IDs to include, if empty all IDs will be included
    :param excluded_ids: list of SOF IDs eo exclude
    :return: (group_dict, group_size) tuple containing a group dictionary and a list of group sizes..
    """
    from tqdm import tqdm
    from math import ceil
    from random import sample

    splits = ['L', 'R'] if split_lr else ['']

    # Generate item keys
    keys = []
    for example in tqdm(dataset, desc='Scanning dataset', unit='example'):
        sof_id = example['id'].numpy()
        visit = example['visit'].numpy()

        if included_ids and sof_id not in included_ids or sof_id in excluded_ids:
            continue
        if visits and visit not in visits:
            continue

        for split in splits:
            key = (sof_id, visit, split)
            keys.append(key)

    if (not num_groups and not max_group_size) or (num_groups and max_group_size):
        raise ValueError("Must set either num_groups or max_group_size")

    # group items
    num_items = len(keys)
    num_groups = num_groups if num_groups else int(ceil(num_items / max_group_size))
    max_group_size = max_group_size if max_group_size else int(ceil(num_items / num_groups))

    group_sizes = [0 for i in range(num_groups)]

    group_dict = {}
    for key in keys:
        available_groups = [i for i in range(num_groups) if group_sizes[i] < max_group_size]

        assert available_groups

        group = sample(available_groups, 1)[0] if randomized else available_groups[0]

        group_sizes[group] += 1
        group_dict[key] = group

    return group_dict, group_sizes



import setuptools
import sof_utils

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as fh:
    requirements = fh.readlines()
requirements = [line.strip('\n') for line in requirements]

setuptools.setup(
    name="sof_utils",
    version=sof_utils.__version__,
    author="Stefan Reinhold",
    description="Small utilities to aid working with the SOF dataset",
    author_email="stefan@sreinhold.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ithron/SOF-Utils",
    packages=setuptools.find_packages(),
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    scripts=[
        'bin/sof-dicom-meta',
        'bin/sof-dicom-corrupted',
        'bin/sof-export-images',
        'bin/sof-convert-labels',
        'bin/sof-convert-tfds',
        'bin/sof-detect-keypoints',
        'bin/sof-keypoint-outliers'
    ],
    install_requires=requirements
)



import tensorflow as tf

def int64_feature(value):
  return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def int64_list_feature(value):
  return tf.train.Feature(int64_list=tf.train.Int64List(value=value))


def bytes_feature(value):
  return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def bytes_list_feature(value):
  return tf.train.Feature(bytes_list=tf.train.BytesList(value=value))


def float_feature(value):
  return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


def float_list_feature(value):
  return tf.train.Feature(float_list=tf.train.FloatList(value=value))